#include <iostream>
#include <fstream>
#include <string>

#include <oclUtils.h>

double walltime ();

void check_cl_err (cl_int err, const char * name)
{
  if (err != 0)
  {
    std::cerr << name << " error: " << err << std::endl;
    exit (1);
  }
}

int main( int argc, const char** argv) 
{
   int N = 500;

   cl_int err;
   char chBuffer[1024];
   cl_uint ciDeviceCount;
   cl_device_id *devices;

   err = clGetPlatformInfo ((cl_platform_id)CL_PLATFORM_NVIDIA, 
     CL_PLATFORM_VERSION, sizeof(chBuffer), chBuffer, NULL);
   check_cl_err (err, "clGetPlatformInfo");

   err = clGetDeviceIDs ((cl_platform_id)CL_PLATFORM_NVIDIA, 
     CL_DEVICE_TYPE_ALL, 0, NULL, &ciDeviceCount);
   check_cl_err (err, "clGetDeviceIDs");

   if (ciDeviceCount == 0) 
   {
     std::cout << "There is no device supporting OpenCL" << std::endl;
     return 0;
   }
   else if (ciDeviceCount == 1)
     std::cout << "There is 1 device supporting OpenCL" << std::endl;
   else
     std::cout << "There are " << ciDeviceCount << " devices supporting OpenCL" << std::endl;

   devices = new cl_device_id [ciDeviceCount];
   err = clGetDeviceIDs ((cl_platform_id)CL_PLATFORM_NVIDIA, 
      CL_DEVICE_TYPE_ALL, ciDeviceCount, devices, &ciDeviceCount);
   check_cl_err (err, "clGetDeviceIDs");
   for (int i=0; i<ciDeviceCount; i++)
   {
     size_t param_value_size_ret;
     cl_uint param_value;

     clGetDeviceInfo (devices[i], CL_DEVICE_VENDOR_ID, sizeof(cl_uint),
                      &param_value, &param_value_size_ret);

     std::cout << "param_value: " << param_value << std::endl;
   }

   cl_context context;
   context = clCreateContext (0, 1, devices, NULL, NULL, &err);
   check_cl_err (err, "clCreateContext");

   cl_command_queue queue;
   queue = clCreateCommandQueue (context, devices[0], 0,
       &err);
   check_cl_err (err, "clCreateCommandQueue");

   cl_mem d_mtx_1;
   d_mtx_1 = clCreateBuffer (context, CL_MEM_READ_ONLY,
       sizeof(float)*N*N, NULL, &err);
   check_cl_err (err, "clCreateBuffer");

   cl_mem d_mtx_2;
   d_mtx_2 = clCreateBuffer (context, CL_MEM_READ_ONLY,
       sizeof(float)*N*N, NULL, &err);
   check_cl_err (err, "clCreateBuffer");

   cl_mem d_mtx_3;
   d_mtx_3 = clCreateBuffer (context, CL_MEM_WRITE_ONLY,
       sizeof(float)*N*N, NULL, &err);
   check_cl_err (err, "clCreateBuffer");

   cl_mem d_N;
   d_N = clCreateBuffer (context, CL_MEM_WRITE_ONLY,
       sizeof(int), NULL, &err);
   check_cl_err (err, "clCreateBuffer");

   float * h_mtx_1, * h_mtx_2;
   h_mtx_1 = new float [N*N];
   h_mtx_2 = new float [N*N]; 

   for (int i=0;i<N;i++)
   {
     for (int j=0;j<N;j++)
     {
       h_mtx_1[(i*N)+j] = i*j;
       h_mtx_2[(i*N)+j] = i+j;
     }
   }

   double wt = walltime();

   err = clEnqueueWriteBuffer (queue, d_mtx_1, CL_TRUE,
       0, sizeof(float)*N*N, (void *)h_mtx_1,
       0, NULL, NULL);
   check_cl_err (err, "clEnqueueWriteBuffer");

   err = clEnqueueWriteBuffer (queue, d_mtx_2, CL_TRUE,
       0, sizeof(float)*N*N, (void *)h_mtx_2,
       0, NULL, NULL);
   check_cl_err (err, "clEnqueueWriteBuffer");

   err = clEnqueueWriteBuffer (queue, d_N, CL_TRUE,
       0, sizeof(int), (void *)&N,
       0, NULL, NULL);
   check_cl_err (err, "clEnqueueWriteBuffer");

   clFinish (queue);

   wt = walltime() - wt;

   std::cout << "Transfer rate: " << 
     ((2*sizeof(float)*N*N)/(1024.0e0*1024.0e0))/wt << " MB/s" << std::endl;

   std::ifstream file("kernel.cl");
   std::string prog(std::istreambuf_iterator<char>(file),
                    (std::istreambuf_iterator<char>()));
   cl_program program;
   const char * ptr = prog.c_str();
   size_t sizeprt = prog.length()+1;
   program = clCreateProgramWithSource (context, 1,
       (const char **)&ptr, &sizeprt, &err);
   check_cl_err (err, "clCreateProgramWithSource");

   err = clBuildProgram (program, 0, NULL, NULL, NULL, NULL);
   check_cl_err (err, "clBuildProgram");
   clUnloadCompiler ();

   cl_kernel kernel;
   kernel = clCreateKernel (program, "mtxsum", &err);
   check_cl_err (err, "clCreateKernel");

   err = clSetKernelArg (kernel, 0, sizeof(cl_mem),
       (const void *) d_mtx_1);
   check_cl_err (err, "clSetKernelArg");

   err = clSetKernelArg (kernel, 1, sizeof(cl_mem),
       (const void *) d_mtx_2);
   check_cl_err (err, "clSetKernelArg");

   err = clSetKernelArg (kernel, 2, sizeof(cl_mem),
       (const void *) d_mtx_3);
   check_cl_err (err, "clSetKernelArg");

   err = clSetKernelArg (kernel, 3, sizeof(int),
       (const void *) &d_N);
   check_cl_err (err, "clSetKernelArg");

   size_t global_work_size[2], local_work_size[2];
   global_work_size[0] = N; global_work_size[1] = N;
   local_work_size[0] = N/2; local_work_size[1] = N/2;
   err = clEnqueueNDRangeKernel (queue, kernel, 2,
       NULL, global_work_size, local_work_size,
       0, NULL, NULL);
   check_cl_err (err, "clEnqueueNDRangeKernel");
   clFinish(queue);

   float * h_mtx_3 = new float [N*N];
   err = clEnqueueReadBuffer (queue, d_mtx_3,
       CL_TRUE, 0,
       sizeof(float)*N*N,
       (void *)h_mtx_3,
       0, NULL, NULL);
   check_cl_err (err, "clEnqueueReadBuffer");

   delete [] h_mtx_1;
   delete [] h_mtx_2;
   delete [] h_mtx_3;

   clReleaseKernel (kernel);
   clReleaseProgram (program);
   clReleaseCommandQueue (queue);
   clReleaseContext (context);
   delete [] devices;

   return 0;
}
